"
COPYRIGHT (c) 2020 LabWare
"
"{ Package: 'stx:goodies/ring/core' }"

"{ NameSpace: Smalltalk }"

RGObject subclass:#RGPackage
	instanceVariableNames:'definedBehaviors extensionMethods tagsForClasses'
	classVariableNames:''
	poolDictionaries:''
	category:'Ring-Core-Kernel'
!

RGPackage comment:'RGPackage is the representation for packages.
A package manages classes, methods and children packages as elements.
A package can also know in which package is defined
'
!

!RGPackage class methodsFor:'documentation'!

copyright
"
COPYRIGHT (c) 2020 LabWare

"
! !

!RGPackage methodsFor:'accessing'!

categories
    
    ^ self definedBehaviors collect: #category as: Set
!

categoryName

    ^ self name 
!

classNames

    ^ self classes collect: [ :each | each name ]
!

classes
    "Retrieves a collection of classes defined in the receiver and classes holding extension methods"
        
    ^self definedClassesOnly, self extendedClasses 
!

definedClassNames

    ^ self definedClassesOnly collect: [ :each | each name ] 
!

definedClasses

    ^ self definedBehaviors
!

definedTraits
    "Retrieves the traits defined in the receiver"

    ^self definedBehaviors select: #isTrait
!

extendedBehaviors
    
    ^ ((self extensionMethods collect: [:each | each parent]) collect: [:behavior |
        behavior isClass 
            ifTrue: [behavior ]
            ifFalse: [ behavior baseClass]] as: IdentitySet) asArray
!

extendedClasses
    
    ^ self extendedBehaviors
!

extendedClassesAndTraits
    
    ^ self extendedBehaviors
!

extendedSelectors
    "Retrieves the names of the methods"
    
    ^ self extensionMethods collect: [ :each | each selector ]
!

extendedTraits
    
    ^ (self extensionMethods collect: [:each | each parent]) select: [:each | each isTrait]
!

extensionMethods

    | allMethods |
    allMethods := IdentitySet new.
    self extensionMethodsDo: [ :each | allMethods add: each].
    ^ allMethods asArray
!

methods
    "Retrieves all the methods defined in the receiver. 
    #methods holds the methods of defined classes"
    
    ^ (self definedBehaviors flatCollect: #methods), self extensionMethods 
!

package

    ^ self
!

traits
    "Retrieves a collection of classes defined in the receiver and classes holding extension methods"
        
    ^self definedTraits, self extendedTraits
! !

!RGPackage methodsFor:'accessing - backend'!

addClassTag: sSymbol

    self backend forPackage addClassTag: sSymbol to: self
!

addDefinedBehavior: anRGBehavior

    self backend forPackage addDefinedBehavior: anRGBehavior to: self
!

addExtensionMethod: anRGMethod 

    self backend forPackage addExtensionMethod: anRGMethod to: self
!

cleanDefinedBehaviors

    self backend forPackage cleanDefinedBehaviorsFor: self
!

cleanExtensionMethods

    self backend forPackage cleanExtensionMethodsFor: self
!

cleanTagsForClasses

    self backend forPackage cleanTagsForClassesFor: self
!

definedBehaviorsDo: aBlock

    self backend forPackage definedBehaviorsFor: self do: aBlock
!

extensionMethodsDo: aBlock

    self backend forPackage extensionMethodsFor: self do: aBlock
!

removeClassTag: aSymbol

    self backend forPackage removeClassTag: aSymbol from: self
!

removeDefinedBehavior: anRGBehavior

    self backend forPackage removeDefinedBehavior: anRGBehavior from: self
!

removeExtensionMethod: anRGMethod 

    self backend forPackage removeExtensionMethod: anRGMethod from: self
!

tagsForClassesDo: aBlock

    self backend forPackage tagsForClassesFor: self do: aBlock
! !

!RGPackage methodsFor:'accessing - model'!

classesTaggedWith: aSymbol

    ^ self definedBehaviors select: [ :each | each isTaggedWith: aSymbol ]
!

tagsForClasses
    "Retrieves the traits defined in the receiver"

    | allTags |
    allTags := IdentitySet new.
    self tagsForClassesDo: [ :each | allTags add: each].
    ^ allTags asArray
! !

!RGPackage methodsFor:'default model values'!

defaultDefinedBehaviors

    ^ Set new
!

defaultExtensionMethods

    ^ Set new
!

defaultTagsForClasses

    ^ Set new
! !

!RGPackage methodsFor:'error handling'!

name: aString

    | oldName |
    
    oldName := self name.
    super name: aString.
    
    self announce: (RPackageRenamed 
            to: self 
            oldName: oldName 
            newName: aString)
! !

!RGPackage methodsFor:'initialization'!

initialize

    super initialize.

    extensionMethods := self unresolvedValue: self defaultExtensionMethods.
    definedBehaviors := self unresolvedValue: self defaultDefinedBehaviors.
    tagsForClasses := self unresolvedValue: self defaultTagsForClasses.
!

initializeUnresolved

    super initializeUnresolved.

    extensionMethods := self unresolvedValue: self defaultDefinedBehaviors.
    definedBehaviors := self unresolvedValue: self defaultExtensionMethods.
    tagsForClasses := self unresolvedValue: self defaultTagsForClasses.
! !

!RGPackage methodsFor:'managing container'!

addoptToParentStub

    self isRingResolved ifFalse: [ ^ self ].
    
    super addoptToParentStub.

    ^ self parent addPackage: self.
!

defaultParentStub

    ^ RGEnvironment new
        yourself.
! !

!RGPackage methodsFor:'printing'!

printOn: aStream
    aStream
        nextPutAll: self class name;
        nextPutAll: '(';
        nextPutAll: self name;
        nextPutAll: ')'
! !

!RGPackage methodsFor:'private - backend interface'!

pvtAddClassTag: aSymbol
    
    tagsForClasses isRingResolved ifFalse: [ 
        self pvtCleanTagsForClasses  ].
    
    tagsForClasses add: aSymbol.	
    
!

pvtAddDefinedBehavior: anRGBehavior

    self environment verifyOwnership: anRGBehavior.
    
    definedBehaviors isRingResolved ifFalse: [ 
        self pvtCleanDefinedBehaviors  ].
    
    definedBehaviors add: anRGBehavior.	
    
!

pvtAddExtensionMethod: aMethodDefinition

    self environment verifyOwnership: aMethodDefinition.
    
    extensionMethods isRingResolved ifFalse: [ 
        self pvtCleanExtensionMethods  ].
    
    extensionMethods add: aMethodDefinition.
    
    "TODO:Announce"
    
    
!

pvtCleanDefinedBehaviors

    definedBehaviors := self defaultDefinedBehaviors.
        
    "TODO:Announce if not empty"
    
    
!

pvtCleanExtensionMethods

    extensionMethods := self defaultExtensionMethods.
        
    "TODO:Announce if not empty"
    
    
!

pvtCleanTagsForClasses

    tagsForClasses := self defaultTagsForClasses.
        
    "TODO:Announce if not empty"
    
    
!

pvtDefinedBehaviorsDo: aBlock

    ^ definedBehaviors value do: aBlock
!

pvtExtensionMethodsDo: aBlock

    ^ extensionMethods value do: aBlock
!

pvtRemoveClassTag: aSymbol
    
    tagsForClasses remove: aSymbol.
    
    "TODO:Announce"
    
    
!

pvtRemoveDefinedBehavior: anRGBehavior

    self environment verifyOwnership: anRGBehavior.
    
    definedBehaviors remove: anRGBehavior.
    
    "TODO:Announce"
    
    
!

pvtRemoveExtensionMethod: anRGMethod

    self environment verifyOwnership: anRGMethod.
    
    extensionMethods remove: anRGMethod.
    
    "TODO:Announce"
    
    
!

pvtResolvableProperties

    ^ super pvtResolvableProperties, { 
        #definedBehaviors -> definedBehaviors. 
        #extensionMethods -> extensionMethods. 
        #tagsForClasses -> tagsForClasses
    }
    
    
!

pvtTagsForClassesDo: aBlock

    ^ tagsForClasses value do: aBlock
! !

!RGPackage methodsFor:'queries - behaviors'!

definedBehaviors

    | allDefinedBehaviors |
    allDefinedBehaviors := IdentitySet new.
    self definedBehaviorsDo: [ :each | allDefinedBehaviors add: each].
    ^ allDefinedBehaviors asArray
    
!

definedClassesOnly

    ^ self definedBehaviors select: #isClass
! !

!RGPackage methodsFor:'resolving'!

makeResolved

    super makeResolved.
    
    definedBehaviors := self definedBehaviors markAsRingResolved.
    extensionMethods := self extensionMethods markAsRingResolved.
    tagsForClasses := self tagsForClasses markAsRingResolved.
! !

!RGPackage methodsFor:'testing'!

isPackage

    ^true
! !

